﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Linq;

namespace QPR.QPRSuite.NCalc {
    public class executesearch : CustomFunctionBase {

        public static string SEARCH_CONFIGURATION_NAMESPACE = "http://www.qpr.com/QPRSuite/ExecuteSearch";
        private XmlNamespaceManager namespaceManager = null;

        public executesearch(FunctionArgs args, IContextManager context)
            : base(args, context, 1, new string[] { "xdocument" }) {
            this.namespaceManager = new XmlNamespaceManager(new NameTable());
            this.namespaceManager.AddNamespace("search", SEARCH_CONFIGURATION_NAMESPACE);
        }

        public override object evaluate() {
            XDocument searchConfiguration = (XDocument)this.parameters[0];
            if (searchConfiguration == null) return new string[0];
            string scopeCombining = this.getAttributeValue(searchConfiguration.Root.Attribute("scopecombining"));
            XElement[] searchScopes = searchConfiguration.XPathSelectElements("/search:executesearch/search:scopepart", this.namespaceManager).ToArray<XElement>();
            XElement generalCriteriaElement = searchConfiguration.XPathSelectElement("/search:executesearch/search:criteria", this.namespaceManager);
            object[] combinedObjects = new string[0];
            for (int i = 0; i < searchScopes.Length; i++) {
                XElement[] attributeElements = searchScopes[i].XPathSelectElements("search:attribute", this.namespaceManager).ToArray<XElement>();
                List<string> attributesAsList = new List<string>();
                List<string> wsAttributes = new List<string>();
                for (int j = 0; j < attributeElements.Length; j++) {
                    attributesAsList.Add(attributeElements[j].Attribute("name").Value.ToLower());
                    wsAttributes.Add(attributeElements[j].Attribute("name").Value.ToLower());
                }
                if (!wsAttributes.Contains("id")) wsAttributes.Add("id");
                string[] attributes = attributesAsList.ToArray<string>();
                XElement specificCriteriaElement = searchScopes[i].XPathSelectElement("search:criteria", this.namespaceManager);
                if (specificCriteriaElement == null) specificCriteriaElement = generalCriteriaElement;
                if (specificCriteriaElement == null) throw new Exception("Search criteria was not found in ExecuteSearch function");
                string webServiceCriteria = this.getWebServiceSearchCriteria(specificCriteriaElement, attributes);
                foreach (string attribute in this.getWsAttributesUsedInCriteria(specificCriteriaElement)) wsAttributes.Add(attribute);
                Dictionary<string, object> arguments = new Dictionary<string, object>();
                arguments.Add("query", searchScopes[i].Attribute("query").Value);
                arguments.Add("criteria", webServiceCriteria);
                arguments.Add("attributes", string.Join(",", wsAttributes.Distinct().ToArray<string>()));
                DataSet results = (DataSet)this.context.getExpressionEngine().evaluateExpressionAsObject("From([query], [criteria], '', [attributes], '')", arguments);
                string expressionEngineCriteria = this.getExpressionEngineSearchCriteria(specificCriteriaElement, attributes);
                arguments = new Dictionary<string, object>();
                arguments.Add("dataset", results);
                arguments.Add("expression", expressionEngineCriteria);
                results = (DataSet)this.context.getExpressionEngine().evaluateExpressionAsObject("Where([dataset], [expression])", arguments);
                arguments = new Dictionary<string, object>();
                string transformation = searchScopes[i].Attribute("transformation") == null ? "" : searchScopes[i].Attribute("transformation").Value;
                arguments = new Dictionary<string, object>();
                arguments.Add("combinedObjects", combinedObjects);
                arguments.Add("dataset", results);
                string scopeCombiningFunction = "";
                if (i == 0 || string.IsNullOrEmpty(scopeCombining) || scopeCombining == "or") { //for the first scope (i==0), combining is always "ArrayUnion" because in the left side there is an empty set
                    scopeCombiningFunction = "ArrayUnion";
                } else {
                    scopeCombiningFunction = "ArrayIntersect";
                }
                if (string.IsNullOrEmpty(transformation)) {
                    combinedObjects = (object[])this.context.getExpressionEngine().evaluateExpressionAsObject(scopeCombiningFunction + "([combinedObjects], DatasetColumnToArray([dataset], 'id'))", arguments);
                } else {
                    combinedObjects = (object[])this.context.getExpressionEngine().evaluateExpressionAsObject(scopeCombiningFunction + "([combinedObjects], QueryObjects('[' + Concatenate(DatasetColumnToArray([dataset], 'id'), ',') + ']." + transformation + "', '', '', 'id', ''))", arguments);
                }
            }
            Dictionary<string, object> finalArguments = new Dictionary<string, object>();
            finalArguments.Add("combinedObjects", combinedObjects);
            return (object[])this.context.getExpressionEngine().evaluateExpressionAsObject("ArrayUnique([combinedObjects])", finalArguments);
        }

        private string getWebServiceSearchCriteria(XElement criteria, string[] searchAttributes) {
            XElement[] criteriaPart = criteria.XPathSelectElements("search:criteria", this.namespaceManager).ToArray<XElement>();
            List<string> criteriaPartAsString = new List<string>();
            for (int i = 0; i < criteriaPart.Length; i++) {
                criteriaPartAsString.Add(getWebServiceSearchCriteria(criteriaPart[i], searchAttributes));
            }
            string criteriaType = criteria.Attribute("type").Value;
            if (criteriaType == "text") {
                if (criteriaPartAsString.Count != 0) throw new Exception("\"text\" type of criteria contains subcriterias in ExecuteSearch function");
                string searchtext = this.getAttributeValue(criteria.Attribute("searchtext"));
                string attribute = this.getAttributeValue(criteria.Attribute("attribute"));
                string compariontype = this.getAttributeValue(criteria.Attribute("compariontype"));
                string matchcase = this.getAttributeValue(criteria.Attribute("matchcase"));
                if (string.IsNullOrEmpty(attribute)) {
                    string[] subcriteria = new string[searchAttributes.Length];
                    for (int i = 0; i < searchAttributes.Length; i++) {
                        subcriteria[i] = getWebServiceSearchCriteriaPart(matchcase, compariontype, searchtext, searchAttributes[i]);
                    }
                    return "(" + string.Join(" OR ", subcriteria) + ")";
                } else {
                    return getWebServiceSearchCriteriaPart(matchcase, compariontype, searchtext, attribute);
                }
            } else if ((new string[] { "and", "or" }).Contains(criteriaType)) {
                return "(" + String.Join(" " + criteriaType.ToUpper() + " ", criteriaPartAsString) + ")";
            } else if (criteriaType == "not") {
                if (criteriaPartAsString.Count != 1) throw new Exception("\"not\" type of criteria contains " + criteriaPartAsString.Count + " subcriterias (only one is allowed) in ExecuteSearch function");
                return "NOT " + criteriaPartAsString[0];
            } else {
                throw new Exception("Unknown criteria type \"" + criteriaType + "\" in ExecuteSearch function");
            }
        }

        private string getWebServiceSearchCriteriaPart(string matchcase, string compariontype, string searchtext, string attribute) {
            searchtext = searchtext.Replace("\"", "\\\"");
            if (matchcase == "true") {
                if (compariontype == "contain" || compariontype == "begin") {
                    return "find(\"" + searchtext + "\", " + attribute + ")";
                } else if (compariontype == "is") {
                    return attribute + "=\"" + searchtext + "\"";
                } else {
                    throw new Exception("Comparison type \"" + compariontype + "\" is unknown in ExecuteSearch function");
                }
            } else { //matchcase == "false"
                if (compariontype == "contain" || compariontype == "begin") {
                    return "find(lower(\"" + searchtext + "\"), lower(" + attribute + "))";
                } else if (compariontype == "is") {
                    return "lower(" + attribute + ")=lower(\"" + searchtext + "\")";
                } else {
                    throw new Exception("Comparison type \"" + compariontype + "\" is unknown in ExecuteSearch function");
                }
            }
        }

        private string getExpressionEngineSearchCriteria(XElement criteria, string[] searchAttributes) {
            XElement[] criteriaPart = criteria.XPathSelectElements("search:criteria", this.namespaceManager).ToArray<XElement>();
            List<string> criteriaPartAsString = new List<string>();
            for (int i = 0; i < criteriaPart.Length; i++) {
                criteriaPartAsString.Add(getExpressionEngineSearchCriteria(criteriaPart[i], searchAttributes));
            }
            string criteriaType = criteria.Attribute("type").Value;
            if (criteriaType == "text") {
                if (criteriaPartAsString.Count != 0) throw new Exception("\"text\" type of criteria contains subcriterias in ExecuteSearch function");
                string searchtext = this.getAttributeValue(criteria.Attribute("searchtext"));
                string attribute = this.getAttributeValue(criteria.Attribute("attribute"));
                string compariontype = this.getAttributeValue(criteria.Attribute("compariontype"));
                string matchcase = this.getAttributeValue(criteria.Attribute("matchcase"));
                if (string.IsNullOrEmpty(attribute)) {
                    string[] subcriteria = new string[searchAttributes.Length];
                    for (int i = 0; i < searchAttributes.Length; i++) {
                        subcriteria[i] = getExpressionEngineSearchCriteriaPart(matchcase, compariontype, searchtext, searchAttributes[i]);
                    }
                    return "(" + string.Join(" or ", subcriteria) + ")";
                } else {
                    return getExpressionEngineSearchCriteriaPart(matchcase, compariontype, searchtext, attribute);
                }
            } else if ((new string[] { "and", "or" }).Contains(criteriaType)) {
                return "(" + String.Join(" " + criteriaType + " ", criteriaPartAsString) + ")";
            } else if (criteriaType == "not") {
                if (criteriaPartAsString.Count != 1) throw new Exception("\"not\" type of criteria contains " + criteriaPartAsString.Count + " subcriterias (only one is allowed) in ExecuteSearch function");
                return "not " + criteriaPartAsString[0];
            } else {
                throw new Exception("Unknown criteria type \"" + criteriaType + "\" in ExecuteSearch function");
            }
        }

        private string getExpressionEngineSearchCriteriaPart(string matchcase, string compariontype, string searchtext, string attribute) {
            searchtext = searchtext.Replace("\\", "\\\\").Replace("'", "\\'");
            if (matchcase == "true") {
                if (compariontype == "contain") {
                    return "Contains([" + attribute + "],'" + searchtext + "')";
                } else if (compariontype == "is") {
                    return "[" + attribute + "]='" + searchtext + "'";
                } else if (compariontype == "begin") {
                    return "IndexOf([" + attribute + "],'" + searchtext + "')=0";
                } else {
                    throw new Exception("Comparison type \"" + compariontype + "\" is unknown in ExecuteSearch function");
                }
            } else { //matchcase == "false"
                if (compariontype == "contain") {
                    return "Contains(ToLower([" + attribute + "]), '" + searchtext.ToLower() + "')";
                } else if (compariontype == "is") {
                    return "ToLower([" + attribute + "])='" + searchtext.ToLower() + "'";
                } else if (compariontype == "begin") {
                    return "IndexOf(ToLower([" + attribute + "]),'" + searchtext.ToLower() + "')=0";
                } else {
                    throw new Exception("Comparison type \"" + compariontype + "\" is unknown in ExecuteSearch function");
                }
            }
        }

        private List<string> getWsAttributesUsedInCriteria(XElement criteria) {
            List<string> attributes = new List<string>();
            string attribute = this.getAttributeValue(criteria.Attribute("attribute"));
            if (!string.IsNullOrEmpty(attribute)) attributes.Add(attribute);
            XElement[] subCriteria = criteria.XPathSelectElements("search:criteria", this.namespaceManager).ToArray<XElement>();
            for (int i = 0; i < subCriteria.Length; i++) {
                List<string> subattributes = getWsAttributesUsedInCriteria(subCriteria[i]);
                foreach (string subattribute in subattributes) attributes.Add(subattribute);
            }
            return attributes;
        }

        private string getAttributeValue(XAttribute attribute) {
            if (attribute == null) return "";
            return attribute.Value;
        }
    }
}