﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using QPR.QPRSuite.AcceleratorTools.QPRWebService;
using System;
using System.Collections.Generic;

namespace QPR.QPRSuite.NCalc {
    public class from : CustomFunctionBase {

        public from(FunctionArgs args, IContextManager context)
            : base(args, context, 4, new string[] { "string", "string", "string", "string", "string" }) {
        }

        public override object evaluate() {
            string options = (this.parameters.Length > 4 && this.parameters[4] != null ? (string)this.parameters[4] : "");
            ResultSet results = this.context.getWebServiceClient().queryObjects((string)this.parameters[0], (string)this.parameters[1], (string)this.parameters[2], (string)this.parameters[3], options);
            string[] columns = this.parseQueriedAttributes();
            if (columns == null || columns.Length == 0) throw new Exception("Dataset contains no columns; some attributes need to be defined.");
            DataSet dataSet = new DataSet(columns);
            if (results == null || results.ChildElements == null) return dataSet;
            for (int i = 0; i < results.ChildElements.Length; i++) {
                object[] rowData = new object[columns.Length];
                Dictionary<string, string> resultRow = new Dictionary<string, string>();
                resultRow["id"] = ((ResultObject)results.ChildElements[i]).Id;
                if (results.ChildElements[i].ChildElements != null) {
                    for (int j = 0; j < results.ChildElements[i].ChildElements.Length; j++) {
                        resultRow[((ValueElement)((ResultObject)results.ChildElements[i]).ChildElements[j]).Name] = ((ValueElement)((ResultObject)results.ChildElements[i]).ChildElements[j]).Value;
                    }
                }
                for (int j = 0; j < columns.Length; j++) {
                    if (resultRow.ContainsKey(columns[j])) {
                        rowData[j] = resultRow[columns[j]];
                    } else {
                        rowData[j] = null;
                    }
                }
                dataSet.addRow(rowData);
            }
            return dataSet;
        }

        private string[] parseQueriedAttributes() {
            string attributesAsString = (string)this.parameters[3];
            attributesAsString = ExpressionEngine.replaceQuotationMarks(attributesAsString);
            int index = 0;
            List<string> attributes = new List<string>();
            while (index < attributesAsString.Length) {
                int parenthesisIndex = attributesAsString.IndexOf("(", index);
                int commaIndex = attributesAsString.IndexOf(",", index);
                if (parenthesisIndex == -1 && commaIndex == -1) {
                    attributes.Add(attributesAsString.Substring(index).Trim());
                    index = attributesAsString.Length;
                } else if ((commaIndex != -1 && commaIndex < parenthesisIndex) || parenthesisIndex == -1) {
                    attributes.Add(attributesAsString.Substring(index, commaIndex - index).Trim());
                    index = commaIndex + 1;
                } else { //parenthesis
                    int parenthesisEndIndex = attributesAsString.IndexOf(")", parenthesisIndex);
                    string parametersPart = attributesAsString.Substring(parenthesisIndex + 1, parenthesisEndIndex - parenthesisIndex - 1);
                    if (parametersPart.Contains("as=\"")) {
                        int attributeNameStartIndex = parametersPart.IndexOf("as=\"") + 4;
                        int attributeNameEndIndex = parametersPart.IndexOf("\"", attributeNameStartIndex);
                        attributes.Add(parametersPart.Substring(attributeNameStartIndex, attributeNameEndIndex - attributeNameStartIndex).Trim());
                    } else {
                        attributes.Add(attributesAsString.Substring(index, parenthesisIndex - index).Trim());
                    }
                    index = parenthesisEndIndex + 2;
                }
            }
            return attributes.ToArray();
        }

    }
}