﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;
using System.Linq;

namespace QPR.QPRSuite.NCalc {

    public class fullouterjoin : CustomFunctionBase {

        public fullouterjoin(FunctionArgs args, IContextManager context)
            : base(args, context, 3, new string[] { "DataSet", "DataSet", "string" }) {
        }

        public override object evaluate() {
            DataSet leftDataSet = (DataSet)this.parameters[0];
            DataSet rightDataSet = (DataSet)this.parameters[1];
            if (leftDataSet == null) return rightDataSet;
            if (rightDataSet == null) return leftDataSet;
            string expression = (string)this.parameters[2];
            if (string.IsNullOrEmpty(expression)) throw new Exception("Expression is empty in FullOuterJoin");
            List<string> newColumns = new List<string>(leftDataSet.getColumns());
            newColumns.AddRange(rightDataSet.getColumns());
            if (newColumns.Distinct().Count() < newColumns.Count()) throw new Exception("Datasets contains same columns in LeftJoin/RightJoin. Columns: " + String.Join(", ", newColumns.ToArray<string>()));
            DataSet newDataSet = new DataSet(newColumns.ToArray());
            Boolean[] leftRowUsed = new Boolean[leftDataSet.rowsCount()];
            Boolean[] rightRowUsed = new Boolean[rightDataSet.rowsCount()];
            for (int leftRow = 0; leftRow < leftDataSet.rowsCount(); leftRow++) {
                for (int rightRow = 0; rightRow < rightDataSet.rowsCount(); rightRow++) {
                    Dictionary<string, object> arguments = new Dictionary<string, object>();
                    for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                        arguments.Add(leftDataSet.getColumnName(i), leftDataSet.getCell(i, leftRow));
                    }
                    for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                        arguments.Add(rightDataSet.getColumnName(i), rightDataSet.getCell(i, rightRow));
                    }
                    if (this.context.getExpressionEngine().evaluateExpressionAsBoolean(expression, arguments)) {
                        object[] newRowData = new object[leftDataSet.columnsCount() + rightDataSet.columnsCount()];
                        for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                            newRowData[i] = leftDataSet.getCell(i, leftRow);
                        }
                        for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                            newRowData[i + leftDataSet.columnsCount()] = rightDataSet.getCell(i, rightRow);
                        }
                        newDataSet.addRow(newRowData);
                        leftRowUsed[rightRow] = true;
                        rightRowUsed[rightRow] = true;
                    }
                }
            }
            for (int leftRow = 0; leftRow < leftDataSet.rowsCount(); leftRow++) {
                if (!leftRowUsed[leftRow]) {
                    object[] newRowData = new object[leftDataSet.columnsCount() + rightDataSet.columnsCount()];
                    for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                        newRowData[i] = leftDataSet.getCell(i, leftRow);
                    }
                    for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                        newRowData[i + leftDataSet.columnsCount()] = null;
                    }
                    newDataSet.addRow(newRowData);
                }
            }
            for (int rightRow = 0; rightRow < rightDataSet.rowsCount(); rightRow++) {
                if (!rightRowUsed[rightRow]) {
                    object[] newRowData = new object[leftDataSet.columnsCount() + rightDataSet.columnsCount()];
                    for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                        newRowData[i] = null;
                    }
                    for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                        newRowData[i + leftDataSet.columnsCount()] = rightDataSet.getCell(i, rightRow);
                    }
                    newDataSet.addRow(newRowData);
                }
            }
            return newDataSet;
        }

    }
}