﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Globalization;
using System.IO;
using System.Net;

namespace QPR.QPRSuite.NCalc {

    public class httpfiledata : CustomFunctionBase {

        private string url;
        private string alternateUrl = null;

        public httpfiledata(FunctionArgs args, IContextManager context)
            : base(args, context, 1, new string[] { "string", "string" }) {
            this.url = Convert.ToString(this.parameters[0], CultureInfo.InvariantCulture);
            if (this.parameters.Length > 1) {
                this.alternateUrl = (string)this.parameters[1];
            }
        }

        public override object evaluate() {
            if (String.IsNullOrEmpty(url)) {
                return null;
            }
            byte[] dataAsByteArray = null;
            try {
                WebRequest request = WebRequest.Create(url);
                request.Credentials = CredentialCache.DefaultCredentials;
                WebResponse response = null;
                try {
                    response = request.GetResponse();
                } catch (Exception e) {
                    if (!e.Message.Contains("404") || String.IsNullOrEmpty(this.alternateUrl)) throw e;
                    request = WebRequest.Create(alternateUrl);
                    request.Credentials = CredentialCache.DefaultCredentials;
                    response = request.GetResponse();
                }
                Stream dataStream = response.GetResponseStream();

                using (MemoryStream ms = new MemoryStream()) {
                    dataStream.CopyTo(ms);
                    dataAsByteArray = ms.ToArray();
                }
                response.Close();
            } catch (Exception e) {
                this.context.getLogger().writeLog("Getting data using http failed. Url=\"" + url + "\", " + e.Message, LogLevel.information);
            }
            return dataAsByteArray;
        }

    }
}