﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;
using System.Linq;

namespace QPR.QPRSuite.NCalc {

    public class innerjoin : CustomFunctionBase {

        public innerjoin(FunctionArgs args, IContextManager context)
            : base(args, context, 3, new string[] { "DataSet", "DataSet", "String" }) {
        }

        public override object evaluate() {
            DataSet leftDataSet = (DataSet)this.parameters[0];
            DataSet rightDataSet = (DataSet)this.parameters[1];
            if (leftDataSet == null || rightDataSet == null) return null;
            string expression = (string)this.parameters[2];
            List<string> newColumns = new List<string>(leftDataSet.getColumns());
            newColumns.AddRange(rightDataSet.getColumns());
            if (newColumns.Distinct().Count() < newColumns.Count()) throw new Exception("Datasets contains same columns in InnerJoin. Columns: " + String.Join(", ", newColumns.ToArray<string>()));
            DataSet newDataSet = new DataSet(newColumns.ToArray());
            for (int leftRow = 0; leftRow < leftDataSet.rowsCount(); leftRow++) {
                for (int rightRow = 0; rightRow < rightDataSet.rowsCount(); rightRow++) {
                    Dictionary<string, object> arguments = new Dictionary<string, object>();
                    for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                        arguments.Add(leftDataSet.getColumnName(i), leftDataSet.getCell(i, leftRow));
                    }
                    for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                        arguments.Add(rightDataSet.getColumnName(i), rightDataSet.getCell(i, rightRow));
                    }
                    if (this.context.getExpressionEngine().evaluateExpressionAsBoolean(expression, arguments)) {
                        object[] newRowData = new object[leftDataSet.columnsCount() + rightDataSet.columnsCount()];
                        for (int i = 0; i < leftDataSet.columnsCount(); i++) {
                            newRowData[i] = leftDataSet.getCell(i, leftRow);
                        }
                        for (int i = 0; i < rightDataSet.columnsCount(); i++) {
                            newRowData[i + leftDataSet.columnsCount()] = rightDataSet.getCell(i, rightRow);
                        }
                        newDataSet.addRow(newRowData);
                    }
                }
            }
            return newDataSet;
        }

    }
}