﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;

namespace QPR.QPRSuite.NCalc {
    public class matrix : CustomFunctionBase {

        private Dictionary<string, Dictionary<string, List<object>>> data = new Dictionary<string, Dictionary<string, List<object>>>();
        private List<string> rows = new List<string>();

        public matrix(FunctionArgs args, IContextManager context)
            : base(args, context, 5, new string[] { "dataset", "string", "string", "string", "string", "string" }) {
        }

        public override object evaluate() {
            DataSet oldDataSet = (DataSet)this.parameters[0];
            if (oldDataSet == null) return null;
            string rowExpression = (string)this.parameters[1];
            if (string.IsNullOrEmpty(rowExpression)) throw new Exception("Row expression is not allowed to be empty in Matrix");
            string columnExpression = (string)this.parameters[2];
            if (string.IsNullOrEmpty(columnExpression)) throw new Exception("Column expression is not allowed to be empty in Matrix");
            string valueExpression = (string)this.parameters[3];
            if (string.IsNullOrEmpty(valueExpression)) throw new Exception("Value expression is not allowed to be empty in Matrix");
            string groupingExpression = (string)this.parameters[4];
            if (string.IsNullOrEmpty(groupingExpression)) throw new Exception("Grouping expression is not allowed to be empty in Matrix");
            string rowColumnName = (string)this.parameters[5];
            if (string.IsNullOrEmpty(rowColumnName)) throw new Exception("Row column name is not allowed to be empty in Matrix");

            for (int i = 0; i < oldDataSet.rowsCount(); i++) {
                object[] oldRow = oldDataSet.getRow(i).getData();
                Dictionary<string, object> arguments = new Dictionary<string, object>();
                for (int j = 0; j < oldRow.Length; j++) {
                    arguments.Add(oldDataSet.getColumnName(j), oldRow[j]);
                }

                object rowExpressionValue = this.context.getExpressionEngine().evaluateExpressionAsObject(rowExpression, arguments);
                if (rowExpressionValue == null) rowExpressionValue = "";
                if (!(rowExpressionValue is string)) throw new Exception("Row expression \"" + rowExpression + "\" was not evaluated to string");

                object columnExpressionValue = this.context.getExpressionEngine().evaluateExpressionAsObject(columnExpression, arguments);
                if (columnExpressionValue == null) columnExpressionValue = "";
                if (!(columnExpressionValue is string)) throw new Exception("Column expression \"" + rowExpression + "\" was not evaluated to string");

                object valueExpressionValue = this.context.getExpressionEngine().evaluateExpressionAsObject(valueExpression, arguments);
                this.addValue((string)rowExpressionValue, (string)columnExpressionValue, valueExpressionValue);
            }
            if (this.data.ContainsKey(rowColumnName)) throw new Exception("Matrics already contains a column with name \"" + rowColumnName + "\". Use different row column name.");
            List<string> newColumns = new List<string>();
            foreach (string key in this.data.Keys) {
                newColumns.Add(key);
            }
            newColumns.Sort();
            newColumns.Insert(0, rowColumnName);
            this.rows.Sort();
            DataSet newDataset = new DataSet(newColumns.ToArray());
            for (int row = 0; row < this.rows.Count; row++) {
                object[] newRow = new object[newColumns.Count];
                newRow[0] = this.rows[row];
                for (int column = 1; column < newColumns.Count; column++) {
                    Dictionary<string, object> arguments = new Dictionary<string, object>();
                    object[] array = null;
                    if (this.data.ContainsKey(newColumns[column]) && this.data[newColumns[column]].ContainsKey(this.rows[row])) {
                        List<object> dataAsArray = this.data[newColumns[column]][this.rows[row]];
                        if (dataAsArray != null) {
                            array = dataAsArray.ToArray();
                        }
                    }
                    arguments.Add("value", array);
                    newRow[column] = this.context.getExpressionEngine().evaluateExpressionAsObject(groupingExpression, arguments);
                }
                newDataset.addRow(newRow);
            }
            return newDataset;
        }

        private void addValue(string row, string column, object value) {
            if (!this.data.ContainsKey(column)) this.data.Add(column, new Dictionary<string, List<object>>());
            if (!this.data[column].ContainsKey(row)) this.data[column].Add(row, new List<object>());
            this.data[column][row].Add(value);
            if (!this.rows.Contains(row)) this.rows.Add(row);
        }

    }
}