﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;

namespace QPR.QPRSuite.NCalc {
    public class split : CustomFunctionBase {

        public split(FunctionArgs args, IContextManager context)
            : base(args, context, 3, new string[] { "dataset", "string", "string" }) {
        }

        public override object evaluate() {
            DataSet oldDataSet = (DataSet)this.parameters[0];
            if (oldDataSet == null) return null;
            string splitColumn = (string)this.parameters[1];
            if (string.IsNullOrEmpty(splitColumn)) throw new Exception("Column name is not allowed to be empty in Split");
            string splitExpression = (string)this.parameters[2];
            if (string.IsNullOrEmpty(splitColumn)) throw new Exception("Split expression is not allowed to be empty in Split");

            List<string> newColumns = new List<string>(oldDataSet.getColumns());
            newColumns.Add(splitColumn);
            DataSet newDataset = new DataSet(newColumns.ToArray());
            for (int i = 0; i < oldDataSet.rowsCount(); i++) {
                object[] oldRow = oldDataSet.getRow(i).getData();
                Dictionary<string, object> arguments = new Dictionary<string, object>();
                for (int j = 0; j < oldRow.Length; j++) {
                    arguments.Add(oldDataSet.getColumnName(j), oldRow[j]);
                }
                object splitExpressionValue = this.context.getExpressionEngine().evaluateExpressionAsObject(splitExpression, arguments);
                if (splitExpressionValue == null) continue;
                if (!splitExpressionValue.GetType().IsArray) throw new Exception("Split expression \"" + splitExpression + "\" was not evaluated to array");
                object[] splitExpressionValueAsArray = (object[])splitExpressionValue;
                for (int j = 0; j < splitExpressionValueAsArray.Length; j++) {
                    object[] newRow = new object[oldDataSet.columnsCount() + 1];
                    for (int k = 0; k < oldRow.Length; k++) {
                        newRow[k] = oldRow[k];
                    }
                    newRow[oldDataSet.columnsCount()] = splitExpressionValueAsArray[j];
                    newDataset.addRow(newRow);
                }
            }
            return newDataset;
        }

    }
}