﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace QPR.QPRSuite.NCalc {
    public class DataSet {

        protected string[] columns;
        protected List<DataSetRow> data;

        public DataSet(string[] columns) {
            this.columns = columns;
            this.data = new List<DataSetRow>();
        }

        public DataSet(string[] columns, List<DataSetRow> data) {
            this.columns = columns;
            this.data = data;
        }

        public string[] getColumns() {
            return this.columns;
        }

        public string getColumnName(int columnIndex) {
            if (columnIndex < 0) throw new Exception("Tried to reference dataset column index " + columnIndex + "(first column is number 0)");
            if (this.columns.Length <= columnIndex) throw new Exception("Dataset has " + this.columns.Length + " columns; tried to reference column " + columnIndex + " (first column is number 0)");
            return this.columns[columnIndex];
        }

        public int rowsCount() {
            return this.data.Count;
        }

        public int columnsCount() {
            return this.columns.Length;
        }

        public DataSetRow getRow(int rowNumber) {
            if (rowNumber < 0) throw new Exception("Tried to reference dataset row index " + rowNumber + "(first row is number 0)");
            if (this.data.Count <= rowNumber) throw new Exception("Dataset has " + this.data.Count + " rows; tried to reference row " + rowNumber + " (first row is number 0)");
            return this.data[rowNumber];
        }

        public void addRow(object[] data) {
            DataSetRow row = new DataSetRow(data, this);
            this.data.Add(row);
        }

        public object getCell(int columnNumber, int rowNumber) {
            return this.getRow(rowNumber).getCell(columnNumber);
        }

        public object getCell(string columnName, int rowNumber) {
            return this.getRow(rowNumber).getCell(columnName);
        }

        public int getColumnNumber(string columnName) {
            for (int i = 0; i < this.columns.Length; i++) {
                if (this.columns[i] == columnName) return i;
            }
            return -1;
        }

        public bool containsSameColumns(DataSet other) {
            if (this.columnsCount() != other.columnsCount()) return false;
            List<string> otherColumns = new List<string>(other.columns);
            foreach (string columnName in this.columns) {
                if (!otherColumns.Contains(columnName)) return false;
            }
            return true;
        }

        public DataSet reorderRows(DataSet other) {
            string[] otherColumns = other.getColumns();
            DataSet reorderedDataSet = new DataSet(otherColumns);
            for (int i = 0; i < this.rowsCount(); i++) {
                object[] reorderedRowData = new object[otherColumns.Length];
                for (int j = 0; j < otherColumns.Length; j++) {
                    int columnNumber = this.getColumnNumber(otherColumns[j]);
                    reorderedRowData[j] = this.getCell(columnNumber, i);
                }
                reorderedDataSet.addRow(reorderedRowData);
            }
            return reorderedDataSet;
        }

        public DataSet intersect(DataSet other) {
            if (!this.containsSameColumns(other)) throw new Exception("Datasets don't contain same columns; intersect is not possible");
            DataSet reorderedOther = other.reorderRows(this);
            DataSet newDataSet = new DataSet(this.columns);
            foreach (DataSetRow row in this.data.Intersect<DataSetRow>(reorderedOther.data)) {
                row.copyRow(newDataSet);
            }
            return newDataSet;
        }

        public DataSet distinct() {
            DataSet newDataSet = new DataSet(this.columns);
            foreach (DataSetRow row in this.data.Distinct<DataSetRow>()) {
                row.copyRow(newDataSet);
            }
            return newDataSet;
        }

        public DataSet copy() {
            DataSet copiedDataSet = new DataSet(this.getColumns());
            foreach (DataSetRow row in this.data) {
                row.copyRow(copiedDataSet);
            }
            return copiedDataSet;
        }

        public Boolean contains(DataSetRow other) {
            return this.data.Contains(other);
        }

        public void sort(DatasetComparer comparer) {
            this.data.Sort(comparer);
        }

        public Dictionary<string, object> getRowAsDictionary(int rowNumber) {
            Dictionary<string, object> result = new Dictionary<string, object>();
            for (int i = 0; i < this.columns.Length; i++) {
                result.Add(this.columns[i], this.getCell(i, rowNumber));
            }
            return result;
        }

    }
}
