﻿using NCalc;
using System;
using System.Collections.Generic;
using System.Globalization;
using QPR.QPRSuite.AcceleratorTools.QPRWebService;
using System.Xml;
using System.Reflection;
using QPR.QPRSuite.AcceleratorTools;

namespace QPR.QPRSuite.NCalc {
    public class ExpressionEngine {

        private IContextManager context;
        private string solutionSpecificCustomFunctionsNamespace;
        private Dictionary<string, Type> typeCache;

        public ExpressionEngine(IContextManager context, string solutionSpecificCustomFunctionsNamespace) {
            this.context = context;
            this.solutionSpecificCustomFunctionsNamespace = solutionSpecificCustomFunctionsNamespace;
            this.typeCache = new Dictionary<string, Type>();
        }

        public void customFunctions(string name, FunctionArgs args) {
            name = name.ToLower();

            //check whether a custom function implementation exists in a separate class
            Type classType = null;
            if (!string.IsNullOrEmpty(solutionSpecificCustomFunctionsNamespace) && this.tryFindType(this.solutionSpecificCustomFunctionsNamespace + "." + name, out classType)) {
            } else {
                classType = Type.GetType("QPR.QPRSuite.NCalc." + name);
            }
            if (classType != null) {
                Object[] constructorArguments = new Object[2];
                constructorArguments[0] = args;
                constructorArguments[1] = this.context;
                try {
                    CustomFunctionBase customFunction = (CustomFunctionBase)Activator.CreateInstance(classType, constructorArguments);
                    args.Result = customFunction.evaluate();
                } catch (Exception e) {
                    throw new Exception("Error in calculating function \"" + name + "\"", e);
                }
                return;
            }

            if (name == "getportalurl") {
                if (args.Parameters.Length < 3) throw new Exception("Function \"GetPortalUrl\" doesn't have enough parameters.");
                args.Result = this.context.getWebServiceClient().getPortalUrl(Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture));
            } else if (name == "queryobjectscount") {
                if (args.Parameters.Length < 3) throw new Exception("Function \"QueryObjectsCount\" doesn't have enough parameters.");
                ResultSet results = this.context.getWebServiceClient().queryObjects(Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), "", "id", Convert.ToString(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture));
                args.Result = (results != null && results.TotalResultsReturned > 0 ? results.TotalResultsReturned : 0);
            } else if (name == "queryobjectssum") {
                if (args.Parameters.Length < 4) throw new Exception("Function \"QueryObjectsSum\" doesn't have enough parameters.");
                ResultSet results = this.context.getWebServiceClient().queryObjects(Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), "", Convert.ToString(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[3].Evaluate(), CultureInfo.InvariantCulture));
                double sum = 0;
                for (int i = 0; results != null && results.ChildElements != null && i < results.ChildElements.Length; i++) {
                    if (results.ChildElements[i].ChildElements.Length != 0) {
                        string valueAsString = ((ValueElement)((ResultObject)results.ChildElements[i]).ChildElements[0]).Value;
                        double valueAsDouble;
                        if (!Double.TryParse(valueAsString, NumberStyles.Any, CultureInfo.InvariantCulture, out valueAsDouble)) {
                            throw new Exception("Value \"" + valueAsString + "\" cannot be converted to numeric in QueryObjectsSum function");
                        }
                        sum += valueAsDouble;
                    }
                }
                args.Result = sum;
            } else if (name == "queryobjectsaverage") {
                this.customFunctions("queryobjectssum", args);
                double sum = (double)args.Result;
                this.customFunctions("queryobjectscount", args);
                int count = (int)args.Result;
                if (count == 0) args.Result = null;
                args.Result = sum / count;
            } else if (name == "substring") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"Substring\" doesn't have enough parameters.");
                if (args.Parameters.Length == 2) {
                    args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Substring(Convert.ToInt32(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture));
                } else {
                    args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Substring(Convert.ToInt32(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), Convert.ToInt32(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture));
                }
            } else if (name == "indexof") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"IndexOf\" doesn't have enough parameters.");
                if (args.Parameters.Length == 2) {
                    args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).IndexOf(Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture));
                } else {
                    args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).IndexOf(Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), Convert.ToInt32(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture));
                }
            } else if (name == "charat") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"CharAt\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate())[Convert.ToInt32(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture)];
            } else if (name == "replace") {
                if (args.Parameters.Length < 3) throw new Exception("Function \"Replace\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Replace(Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture), Convert.ToString(args.Parameters[2].Evaluate(), CultureInfo.InvariantCulture));
            } else if (name == "trim") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"Trim\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Trim();
            } else if (name == "tolower") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"ToLower\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).ToLower();
            } else if (name == "toupper") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"ToUpper\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).ToUpper();
            } else if (name == "length") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"Length\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Length;
            } else if (name == "contains") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"Contains\" doesn't have enough parameters.");
                args.Result = Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture).Contains(Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture));
            } else if (name == "stringtodate") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"StringToDate\" doesn't have enough parameters.");
                object parameter1 = args.Parameters[0].Evaluate();
                if (!(parameter1 is String)) throw new Exception("The first parameter must be string in function \"StringToDate\".");
                string stringToConvert = (string)parameter1;
                if (string.IsNullOrEmpty(stringToConvert)) {
                    args.Result = null;
                } else if (args.Parameters.Length == 1) {
                    args.Result = XmlConvert.ToDateTime(stringToConvert, XmlDateTimeSerializationMode.Local);
                } else {
                    string dateFormat = Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture);
                    DateTime dateAsDate;
                    if (!DateTime.TryParseExact(stringToConvert, dateFormat, CultureInfo.InvariantCulture, DateTimeStyles.None, out dateAsDate)) {
                        throw new Exception("Value \"" + stringToConvert + "\" cannot be converted to date in StringToDate function");
                    }
                    args.Result = dateAsDate;
                }
            } else if (name == "datetostring") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"DateToString\" doesn't have enough parameters.");
                object dateToConvert = args.Parameters[0].Evaluate();
                if (!(dateToConvert is DateTime)) throw new Exception("The first parameter must be datetime in function \"DateToString\".");
                args.Result = ((DateTime)dateToConvert).ToString(Convert.ToString(args.Parameters[1].Evaluate(), CultureInfo.InvariantCulture));
            } else if (name == "addmonths") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"AddMonths\" doesn't have enough parameters.");
                object date = args.Parameters[0].Evaluate();
                object months = args.Parameters[1].Evaluate();
                if (!(date is DateTime)) throw new Exception("The first parameter must be datetetime in function \"AddMilliseconds\".");
                args.Result = ((DateTime)date).AddMonths(Convert.ToInt32(months, CultureInfo.InvariantCulture));
            } else if (name == "addmilliseconds") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"AddMilliseconds\" doesn't have enough parameters.");
                object date = args.Parameters[0].Evaluate();
                object milliseconds = args.Parameters[1].Evaluate();
                if (!(date is DateTime)) throw new Exception("The first parameter must be datetetime in function \"AddMilliseconds\".");
                args.Result = ((DateTime)date).AddMilliseconds(Convert.ToDouble(milliseconds, CultureInfo.InvariantCulture));
            } else if (name == "comparedates") {
                if (args.Parameters.Length < 2) throw new Exception("Function \"CompareDates\" doesn't have enough parameters.");
                object date1 = args.Parameters[0].Evaluate();
                object date2 = args.Parameters[1].Evaluate();
                if (!(date1 is DateTime)) throw new Exception("The first parameter must be datetetime in function \"AddMilliseconds\".");
                if (!(date2 is DateTime)) throw new Exception("The second parameter must be datetetime in function \"AddMilliseconds\".");
                args.Result = ((DateTime)date1).CompareTo((DateTime)date2);
            } else if (name == "currentdatetime") {
                args.Result = DateTime.Now;
            } else if (name == "getticks") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"GetDateTicks\" doesn't have enough parameters.");
                object dateValue = args.Parameters[0].Evaluate();
                if (!(dateValue is DateTime)) throw new Exception("The first parameter must be datetetime in function \"GetDateTicks\".");
                args.Result = ((DateTime)dateValue).Ticks;
            } else if (name == "getdate") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"GetDateComponent\" doesn't have enough parameters.");
                object dateValue = args.Parameters[0].Evaluate();
                if (!(dateValue is DateTime)) throw new Exception("The first parameter must be datetetime in function \"GetDateComponent\".");
                args.Result = ((DateTime)dateValue).Date;
            } else if (name == "getyear") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"GetYear\" doesn't have enough parameters.");
                object dateValue = args.Parameters[0].Evaluate();
                if (!(dateValue is DateTime)) throw new Exception("The first parameter must be datetetime in function \"GetYear\".");
                args.Result = ((DateTime)dateValue).Year;
            } else if (name == "getmonth") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"GetMonth\" doesn't have enough parameters.");
                object dateValue = args.Parameters[0].Evaluate();
                if (!(dateValue is DateTime)) throw new Exception("The first parameter must be datetetime in function \"GetMonth\".");
                args.Result = ((DateTime)dateValue).Year;
            } else if (name == "getday") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"GetDay\" doesn't have enough parameters.");
                object dateValue = args.Parameters[0].Evaluate();
                if (!(dateValue is DateTime)) throw new Exception("The first parameter must be datetetime in function \"GetDay\".");
                args.Result = ((DateTime)dateValue).Day;
            } else if (name == "applicationversion") {
                args.Result = this.context.getApplicationVersion();
            } else if (name == "raiseerror") {
                if (args.Parameters.Length < 1) throw new Exception("Function \"RaiseError\" doesn't have enough parameters.");
                throw new Exception(Convert.ToString(args.Parameters[0].Evaluate(), CultureInfo.InvariantCulture));
            }
        }

        //evaluates NCalc expression where there are expression parameters
        public object evaluateExpressionAsObject(string inputString, Dictionary<string, object> parameters = null) {
            if (String.IsNullOrEmpty(inputString)) return inputString;
            inputString = ExpressionEngine.replaceQuotationMarks(inputString);
            Expression expression = new Expression(inputString);
            expression.EvaluateFunction += customFunctions;
            if (parameters != null) {
                foreach (string parameter in parameters.Keys) {
                    expression.Parameters[parameter] = parameters[parameter];
                }
            }
            if (expression.HasErrors()) throw new Exception("Syntax error in expression \"" + inputString + "\". Description: " + expression.Error);
            try {
                return expression.Evaluate();
            } catch (EvaluationException e) {
                throw new Exception("Error when evaluating expression \"" + inputString + "\". Description: " + expression.Error, e);
            } catch (ArgumentException e) {
                throw new Exception("Error in arguments for expression \"" + inputString + "\". Note that it's not possible to use reserved names for arguments.", e);
            }
        }

        //evaluates NCalc expression where there are expression parameters
        public object evaluateExpression(Expression expression, Dictionary<string, object> parameters = null) {
            if (expression == null) return null;
            if (parameters != null) {
                foreach (string parameter in parameters.Keys) {
                    expression.Parameters[parameter] = parameters[parameter];
                }
            }
            if (expression.HasErrors()) throw new Exception("Syntax error in expression \"" + expression.ParsedExpression + "\". Description: " + expression.Error);
            try {
                return expression.Evaluate();
            } catch (EvaluationException e) {
                throw new Exception("Evaluation error in expression \"" + expression.ParsedExpression + "\". Description: " + expression.Error, e);
            } catch (ArgumentException e) {
                throw new Exception("Error in arguments for expression \"" + expression.ParsedExpression + "\". Note that it's not possible to use reserved names for arguments.", e);
            }
        }

        public string evaluateExpressionAsString(string expression, Dictionary<string, object> parameters = null) {
            object result = this.evaluateExpressionAsObject(expression, parameters);
            if (!(result is string)) throw new Exception("Expression was not evaluated as string. Expression: " + expression);
            return (string)result;
        }

        public bool evaluateExpressionAsBoolean(string expression, Dictionary<string, object> parameters = null) {
            object result = this.evaluateExpressionAsObject(expression, parameters);
            if (!(result is bool)) throw new Exception("Expression was not evaluated as boolean. Expression: " + expression);
            return (bool)result;
        }

        public Expression convertToExpression(string inputString) {
            if (inputString == null) return null;
            inputString = ExpressionEngine.replaceQuotationMarks(inputString);
            Expression expression = new Expression(inputString);
            expression.EvaluateFunction += customFunctions;
            return expression;
        }

        //replaces "exotic" quotation marks with valid ones for QPR Web Service (MS Word usually uses these exotic quotation marks)
        public static string replaceQuotationMarks(string input) {
            if (input == null) return null;
            return input.Replace('’', '\'').Replace('‘', '\'').Replace('“', '"').Replace('”', '"');
        }

        public bool tryFindType(string typeName, out Type t) {
            lock (typeCache) {
                if (!typeCache.TryGetValue(typeName, out t)) {
                    foreach (Assembly a in AppDomain.CurrentDomain.GetAssemblies()) {
                        t = a.GetType(typeName);
                        if (t != null)
                            break;
                    }
                    typeCache[typeName] = t;
                }
            }
            return t != null;
        }

    }
}
